//
// Main
//

type Func = (...args: any[]) => any

export interface Cache<K, V> {
  create: CacheCreateFunc<K, V>
}

interface CacheCreateFunc<K, V> {
  (): {
    get(key: K): V
    set(key: K, value: V): void
    has(key: K): boolean
  }
}

export type Serializer = (args: any[]) => string

export interface Options<F extends Func> {
  cache?: Cache<string, ReturnType<F>>
  serializer?: Serializer
  strategy?: MemoizeFunc
}

export interface MemoizeFunc {
  <F extends Func>(fn: F, options?: Options<F>): F
}
const memoize: MemoizeFunc = (fn, options) => {
  const cache = options && options.cache ? options.cache : cacheDefault

  const serializer =
    options && options.serializer ? options.serializer : serializerDefault

  const strategy =
    options && options.strategy ? options.strategy : strategyDefault

  return strategy(fn, {
    cache: cache,
    serializer: serializer,
  })
}

//
// Strategy
//

function isPrimitive(value: any): boolean {
  return (
    value == null || typeof value === 'number' || typeof value === 'boolean'
  ) // || typeof value === "string" 'unsafe' primitive for our needs
}

function monadic(fn, cache, serializer, arg) {
  const cacheKey = isPrimitive(arg) ? arg : serializer(arg)

  let computedValue = cache.get(cacheKey)
  if (typeof computedValue === 'undefined') {
    computedValue = fn.call(this, arg)
    cache.set(cacheKey, computedValue)
  }

  return computedValue
}

function variadic(fn, cache, serializer) {
  const args = Array.prototype.slice.call(arguments, 3)
  const cacheKey = serializer(args)

  let computedValue = cache.get(cacheKey)
  if (typeof computedValue === 'undefined') {
    computedValue = fn.apply(this, args)
    cache.set(cacheKey, computedValue)
  }

  return computedValue
}

function assemble(fn, context, strategy, cache, serialize) {
  return strategy.bind(context, fn, cache, serialize)
}

function strategyDefault(fn, options) {
  const strategy = fn.length === 1 ? monadic : variadic

  return assemble(
    fn,
    this,
    strategy,
    options.cache.create(),
    options.serializer
  )
}

function strategyVariadic(fn, options) {
  return assemble(
    fn,
    this,
    variadic,
    options.cache.create(),
    options.serializer
  )
}

function strategyMonadic(fn, options) {
  return assemble(fn, this, monadic, options.cache.create(), options.serializer)
}

//
// Serializer
//

function serializerDefault(): string {
  return JSON.stringify(arguments)
}

//
// Cache
//

function ObjectWithoutPrototypeCache() {
  this.cache = Object.create(null) as Record<string, any>
}

ObjectWithoutPrototypeCache.prototype.has = function (key: string) {
  return key in this.cache
}

ObjectWithoutPrototypeCache.prototype.get = function (key: string) {
  return this.cache[key]
}

ObjectWithoutPrototypeCache.prototype.set = function <T>(
  key: string,
  value: T
): void {
  this.cache[key] = value
}

var cacheDefault = {
  create: function create() {
    return new ObjectWithoutPrototypeCache()
  },
}

//
// API
//

export const strategies = {
  variadic: strategyVariadic,
  monadic: strategyMonadic,
}

export default memoize
